from typing import Tuple, Callable, Union, Optional
import jax
import jax.numpy as jnp
import jax.random as jr
import equinox as eqx
from jaxtyping import Key, Array


class Linear(eqx.Module):
    weight: Array
    bias: Array

    def __init__(
        self, 
        in_size: int, 
        out_size: int, 
        *, 
        key: Key
    ):
        lim = jnp.sqrt(1. / (in_size + 1.))
        key_w, _ = jr.split(key)
        self.weight = jr.truncated_normal(
            key_w, shape=(out_size, in_size), lower=-2., upper=2.
        ) * lim
        self.bias = jnp.zeros((out_size,))

    def __call__(self, x: Array) -> Array:
        return self.weight @ x + self.bias


class ResidualNetwork(eqx.Module):
    _in: Linear
    layers: Tuple[Linear]
    dropouts: Tuple[eqx.nn.Dropout]
    _out: Linear
    activation: Callable
    q_dim: Optional[int] = None
    a_dim: Optional[int] = None
    t1: float

    def __init__(
        self, 
        in_size: int, 
        width_size: int, 
        depth: Optional[int], 
        q_dim: Optional[int] = None,
        a_dim: Optional[int] = None, 
        activation: Callable = jax.nn.tanh,
        dropout_p: float = 0.,
        t1: float = 1.,
        *, 
        key: Key
    ):
        """
            Parameters:
            -----------
            `in_size` : `int`
                The size of the input features to the network.
            
            `width_size` : `int`
                The width (number of units) of each hidden layer in the network.
            
            `depth` : `Optional[int]`, default: `None`
                The number of residual layers in the network. 
            
            `q_dim` : `Optional[int]`, default: `None`
                The number of elements in the optional conditioning input.
            
            `a_dim` : `Optional[int]`, default: `None`
                The number of parameters in the optional conditioning.
            
            `activation` : `Callable`, default: `jax.nn.tanh`
                The activation function applied after each layer. Defaults to the hyperbolic tangent (`tanh`).
            
            `dropout_p` : `float`, default: `0.`
                The probability of dropping out units in the dropout layers. Defaults to `0` (no dropout).

            `t1` : float, default: `1.0`
                Default maximum time of diffusion process. Scales input times.
            
            `key` : `Key`
                JAX random key used for initialization.
        """

        in_key, *net_keys, out_key = jr.split(key, 2 + depth)

        _in_size = in_size + 1 # TODO: time embedding
        if q_dim is not None:
            _in_size += q_dim
        if a_dim is not None:
            _in_size += a_dim

        _width_size = width_size + 1 # TODO: time embedding
        if q_dim is not None:
            _width_size += q_dim
        if a_dim is not None:
            _width_size += a_dim

        self._in = Linear(_in_size,width_size, key=in_key)
        layers = [
            Linear(_width_size, width_size, key=_key)
            for _key in net_keys 
        ]
        self._out = Linear(width_size, in_size, key=out_key)
        self.layers = tuple(layers)

        dropouts = [
            eqx.nn.Dropout(p=dropout_p) for _ in layers
        ]
        self.dropouts = tuple(dropouts)

        self.activation = activation
        self.q_dim = q_dim
        self.a_dim = a_dim
        self.t1 = t1
    
    def __call__(
        self, 
        t: Union[float, Array], 
        x: Array, 
        q: Optional[Array], 
        a: Optional[Array], 
        *, 
        key: Key = None
    ) -> Array:
        t = jnp.atleast_1d(t / self.t1)

        _qa = []
        if q is not None and self.q_dim is not None:
            _qa.append(q)
        if a is not None and self.a_dim is not None:
            _qa.append(a)

        xqat = jnp.concatenate([x, t] + _qa)
        
        h0 = self._in(xqat)
        h = h0
        for l, d in zip(self.layers, self.dropouts):
            # Condition on time at each layer
            hqat = jnp.concatenate([h, t] + _qa)

            h = l(hqat)
            h = d(h, key=key)
            h = self.activation(h)
        o = self._out(h0 + h)
        return o


import jax.numpy as jnp
import jax.random as jr 
from sklearn.datasets import make_moons

from .utils import ScalerDataset, InMemoryDataLoader, Scaler, Normer


def key_to_seed(key):
    return int(jnp.asarray(jr.key_data(key)).sum())

def moons(key):
    key_train, key_valid = jr.split(key)

    data_shape = (2,)
    context_shape = None 
    parameter_dim = 1

    Xt, Yt = make_moons(
        40_000, noise=0.05, random_state=key_to_seed(key_train)
    )
    Xv, Yv = make_moons(
        40_000, noise=0.05, random_state=key_to_seed(key_valid)
    )
    Yt = Yt[:, jnp.newaxis].astype(jnp.float32)
    Yv = Yv[:, jnp.newaxis].astype(jnp.float32)

    process_fn = Normer(Xt.mean(), Xt.std())
    
    train_dataloader = InMemoryDataLoader(
        X=jnp.asarray(Xt), Q=jnp.asarray(Yt), A=None, process_fn=process_fn, key=key_train
    )
    valid_dataloader = InMemoryDataLoader(
        X=jnp.asarray(Xv), Q=jnp.asarray(Yv), A=None, process_fn=process_fn, key=key_valid
    )

    def label_fn(key, n):
        Q = jr.choice(key, jnp.array([0., 1.]), (n,))[:, jnp.newaxis]
        A = None
        return Q, A

    return ScalerDataset(
        name="moons",
        train_dataloader=train_dataloader,
        valid_dataloader=valid_dataloader,
        data_shape=data_shape,
        context_shape=context_shape,
        parameter_dim=parameter_dim,
        process_fn=process_fn, # Scaler(x_min=Xt.min(), x_max=Xt.max()),
        label_fn=label_fn
    )


import os
import jax.random as jr 
import jax.numpy as jnp
from jaxtyping import Key
from torchvision import transforms, datasets

from .utils import Scaler, Normer, ScalerDataset, TorchDataLoader, InMemoryDataLoader


def convert_torch_to_in_memory(dataset):
    # Convert torch cifar10 dataset to in-memory
    data = jnp.asarray(dataset.data)
    data = data.transpose(0, 3, 1, 2).astype(jnp.float32)
    data = data / data.max()
    targets = jnp.asarray(dataset.targets).astype(jnp.float32)
    targets = targets[:, jnp.newaxis]
    return data, targets


def cifar10(path: str, key: Key, *, in_memory: bool = True) -> ScalerDataset:

    key_train, key_valid = jr.split(key)

    n_pix = 32 # Native resolution for CIFAR10 
    data_shape = (3, n_pix, n_pix)
    context_shape = None
    parameter_dim = 1
    n_classes = 10

    scaler = Normer()

    train_transform = transforms.Compose(
        [
            transforms.Resize((n_pix, n_pix)),
            transforms.RandomHorizontalFlip(),
            transforms.ToTensor(), 
            transforms.Lambda(scaler.forward) 
        ]
    )
    valid_transform = transforms.Compose(
        [
            transforms.Resize((n_pix, n_pix)),
            transforms.ToTensor(),
            transforms.Lambda(scaler.forward)
        ]
    )
    train_dataset = datasets.CIFAR10(
        os.path.join(path, "datasets/cifar10/"),
        train=True, 
        download=True, 
        transform=train_transform,
        target_transform=transforms.Lambda(lambda x: x.float())
    )
    valid_dataset = datasets.CIFAR10(
        os.path.join(path, "datasets/cifar10/"),
        train=False, 
        download=True, 
        transform=valid_transform,
        target_transform=transforms.Lambda(lambda x: x.float())
    )

    if in_memory:
        Xt, At = convert_torch_to_in_memory(train_dataset) 
        Xv, Av = convert_torch_to_in_memory(valid_dataset) 

        At = At.astype(jnp.float32)
        Av = Av.astype(jnp.float32)

        process_fn = Normer(x_mean=Xt.mean(), x_std=Xt.std())

        train_dataloader = InMemoryDataLoader(
            X=Xt, A=At, process_fn=process_fn, key=key_train) 
        valid_dataloader = InMemoryDataLoader(
            X=Xv, A=Av, process_fn=process_fn, key=key_valid
        ) 
    else:
        process_fn = Scaler(x_min=0., x_max=1.)

        train_dataloader = TorchDataLoader(
            train_dataset, 
            data_shape=data_shape, 
            context_shape=context_shape,
            parameter_dim=parameter_dim, 
            key=key_train
        )
        valid_dataloader = TorchDataLoader(
            valid_dataset, 
            data_shape=data_shape, 
            context_shape=context_shape,
            parameter_dim=parameter_dim, 
            key=key_valid
        )

    def label_fn(key, n):
        Q = None
        A = jr.choice(key, jnp.arange(n_classes), (n,))
        A = A[:, jnp.newaxis].astype(jnp.float32)
        return Q, A

    return ScalerDataset(
        name="cifar10",
        train_dataloader=train_dataloader,
        valid_dataloader=valid_dataloader,
        data_shape=data_shape,
        parameter_dim=parameter_dim,
        context_shape=None,
        process_fn=process_fn, 
        label_fn=label_fn
    )


from ml_collections import ConfigDict

from ._sde import SDE, default_weight_fn
from ._ve import VESDE
from ._vp import VPSDE
from ._subvp import SubVPSDE


def get_sde(config: ConfigDict) -> SDE:
    assert config.sde in ["VE", "VP", "SubVP"]

    name = config.sde + "SDE"

    sdes = [VESDE, VPSDE, SubVPSDE]
    sde = sdes[[sde.__name__ for sde in sdes].index(name)]

    if hasattr(config, "weight_fn"):
        weight_fn = config.weight_fn
    else:
        weight_fn = default_weight_fn

    if config.sde in ["VP", "SubVP"]:
        _sde = sde(
            beta_integral_fn=config.beta_integral,
            dt=config.dt,
            t0=config.t0, 
            t1=config.t1,
            weight_fn=weight_fn
        )
    if config.sde in ["VE"]:
        _sde = sde(
            sigma_fn=config.sigma_fn,
            dt=config.dt,
            t0=config.t0, 
            t1=config.t1,
            weight_fn=weight_fn
        )
    return _sde


from typing import Sequence, Callable, Optional
import diffrax as dfx  
import jax
import jax.numpy as jnp
import jax.random as jr
import equinox as eqx
from jaxtyping import PRNGKeyArray, Array, Float, jaxtyped
from beartype import beartype as typechecker

from .sde import SDE
from ._ode import get_solver

"""
    Implements sampling from the reverse-time stochastic differential equation (SDE) using either an ODE solver or
    Euler-Maruyama method. 
    - This module provides functions to generate samples from a trained score-based generative model
      by solving the reverse-time SDE defined by the model.
"""


@jaxtyped(typechecker=typechecker)
@eqx.filter_jit
def single_ode_sample_fn(
    model: eqx.Module, 
    sde: SDE, 
    data_shape: Sequence[int], 
    key: PRNGKeyArray,
    q: Optional[Float[Array, "..."]] = None,
    a: Optional[Float[Array, "..."]] = None,
    solver: Optional[dfx.AbstractSolver] = None
) -> Array:
    """
        Solves the reverse-time ordinary differential equation (ODE) to generate a sample from the prior distribution 
        by solving the initial value problem (IVP) with the provided model and stochastic differential equation (SDE).

        Parameters:
        -----------
        `model` : `eqx.Module`
            The trained model, typically a score-based generative model, used for reverse-time sampling.
        
        `sde` : `SDE`
            The stochastic differential equation (SDE) defining both the forward and reverse diffusion dynamics.
        
        `data_shape` : `Sequence[int]`
            Shape of the data to be sampled, used for constructing the prior sample.
        
        `key` : `PRNGKeyArray`
            A JAX random key used to generate the prior sample.
        
        `q` : `Optional[Array]`, default: `None`
            Optional conditioning variable `q` for the ODE, if applicable.
        
        `a` : `Optional[Array]`, default: `None`
            Optional conditioning variable `a` for the ODE, if applicable.
        
        `solver` : `Optional[dfx.AbstractSolver]`, default: `None`
            The differential equation solver to be used for solving the reverse-time ODE. If `None`, a default solver is used.

        Returns:
        --------
        `Array`
            The generated sample at time `t0`, obtained by solving the reverse-time ODE from the prior sample at `t1`.

        Notes:
        ------
        - The function uses the reverse SDE (with probability flow) to define the drift term for the ODE, 
        which corresponds to the reverse diffusion process.
        - The ODE is solved from `t1` (the endpoint of the forward process) to `t0` (the initial point).
        - If `q` and `a` are provided, they condition the reverse-time dynamics during the sampling process.
        - The reverse-time SDE is solved by `diffeqsolve` from the `diffrax` library, with `ODETerm` representing 
        the reverse dynamics.

        Example:
        --------
        ```python
        sampled_data = single_ode_sample_fn(
            model, sde, data_shape=(3, 32, 32), key=jr.PRNGKey(0)
        )
        ```
    """

    model = eqx.nn.inference_mode(model, True)

    reverse_sde = sde.reverse(model, probability_flow=True)

    def ode(t, y, args):
        """
            dx = [f(x, t) - 0.5 * g^2(t) * score(x, t, q)] * dt
        """
        (q, a) = args
        drift, _ = reverse_sde.sde(y, t, q, a)
        return drift

    term = dfx.ODETerm(ode)
    solver = solver if solver is not None else get_solver() 
    y1 = sde.prior_sample(key, data_shape) 

    # Reverse time, solve from t1 to t0
    sol = dfx.diffeqsolve(
        term, solver, sde.t1, sde.t0, -sde.dt, y1, (q, a)
    )
    return sol.ys[0]


@jaxtyped(typechecker=typechecker)
@eqx.filter_jit
def single_eu_sample_fn(
    model: eqx.Module, 
    sde: SDE, 
    data_shape: Sequence[int], 
    key: PRNGKeyArray, 
    q: Optional[Float[Array, "..."]] = None,
    a: Optional[Float[Array, "..."]] = None,
    T_sample: int = 1_000
) -> Array:
    """
        Implements the Euler-Maruyama sampler for solving the reverse-time stochastic differential equation (SDE) 
        to generate a sample from the prior distribution.

        Parameters:
        -----------
        `model` : `eqx.Module`
            The trained model, typically a score-based generative model, used for reverse-time sampling.
        
        `sde` : `SDE`
            The stochastic differential equation (SDE) defining both the forward and reverse diffusion dynamics.
        
        `data_shape` : `Sequence[int]`
            Shape of the data to be sampled, used for constructing the prior sample.
        
        `key` : `PRNGKeyArray`
            A JAX random key used to generate the prior sample and noise at each time step.
        
        `q` : `Optional[Array]`, default: `None`
            Optional conditioning variable `q` for the SDE, if applicable.
        
        `a` : `Optional[Array]`, default: `None`
            Optional conditioning variable `a` for the SDE, if applicable.
        
        `T_sample` : `int`, default: `1_000`
            The number of time steps used for the Euler-Maruyama discretization. Higher values give more accurate 
            approximations of the reverse SDE.

        Returns:
        --------
        `Array`
            The generated sample at time `t0`, obtained by solving the reverse-time SDE using Euler-Maruyama discretization.

        Notes:
        ------
        - The function simulates the reverse diffusion process starting from a prior sample `x_T` at time `t1`, 
          and evolves it to `t0` using the Euler-Maruyama method.
        - At each time step `i`, the function applies the Euler-Maruyama update rule: 
          `x <- x + [f(x, t) - g^2(t) * score(x, t, q)] * dt + g(t) * sqrt(dt) * eps_t`, 
          where `f(x, t)` is the drift term, `g(t)` is the diffusion coefficient, and `eps_t` is a random noise term.
        - The reverse-time SDE is defined with `probability_flow=False`, meaning it includes both drift and diffusion components.

        Example:
        --------
        ```python
        sampled_data = single_eu_sample_fn(
            model, sde, data_shape=(32, 32, 3), key=jr.PRNGKey(0)
        )
        ```
    """
    
    model = eqx.nn.inference_mode(model, True)

    time_steps = jnp.linspace(sde.t1, sde.t0, T_sample) # Reversed time
    step_size = (sde.t1 - sde.t0) / T_sample 

    x_T = sde.prior_sample(key, data_shape) 
    reverse_sde = sde.reverse(model, probability_flow=False)

    def marginal(i, val):
        """
            Euler Maryuma iteration:
            dx <- [f(x, t) - g^2(t) * score(x, t, q)] * dt + g(t) * sqrt(dt) * eps_t
            x <- x + dx
            t <- t + dt
        """
        (mean_x, x) = val
        t = time_steps[i]
        
        key_eps = jr.fold_in(key, i) 

        eps_t = jr.normal(key_eps, data_shape)
        drift, diffusion = reverse_sde.sde(x, t, q, a)
        mean_x = x - drift * step_size # mu_x = x + drift * -step

        # x = [f(x, t) - g^2(t) * score(x, t, q)] * dt + g(t) * sqrt(dt) * eps_t
        x = mean_x + diffusion * jnp.sqrt(step_size) * eps_t

        return mean_x, x

    mean_x, x = jax.lax.fori_loop(
        0, T_sample, marginal, init_val=(jnp.zeros_like(x_T), x_T)
    )

    # Do not include any noise in the last sampling step.
    return mean_x


def get_eu_sample_fn(
    model: eqx.Module, 
    sde: SDE, 
    data_shape: Sequence[int], 
    T_sample: int = 1_000
) -> Callable:
    """
        Returns a callable function that implements Euler-Maruyama sampling of the reverse-time stochastic differential 
        equation (SDE) for generating data samples.

        Parameters:
        -----------
        `model` : `eqx.Module`
            The trained score network model used for reverse-time sampling.
        
        `sde` : `SDE`
            The stochastic differential equation (SDE) defining both the forward diffusion dynamics.
        
        `data_shape` : `Sequence[int]`
            Shape of the data to be sampled, used for constructing the prior sample.
        
        `T_sample` : `int`, default: `1_000`
            The number of time steps used for the Euler-Maruyama discretization. Higher values give more accurate 
            approximations of the reverse SDE.

        Returns:
        --------
        `Callable`
            A function `_eu_sample_fn` which generates a data sample given a random key `key`, and optional 
            conditioning variables `q` and `a`. It has the following signature:
            
            ```python
            def _eu_sample_fn(key: PRNGKeyArray, q: Optional[Array], a: Optional[Array]) -> Array
            ```

            This function uses `single_eu_sample_fn` to generate a sample at time `t0`.

        Example:
        --------
        ```python
        eu_sampler = get_eu_sample_fn(model, sde, data_shape=(3, 32, 32))
        sampled_data = eu_sampler(jr.PRNGKey(0), q=None, a=None)
        ```

        Notes:
        ------
        - The returned function calls `single_eu_sample_fn` to simulate the reverse diffusion process starting from 
          a prior sample `x_T` at time `t1`, and evolves it to `t0` using the Euler-Maruyama method.
        - The sample function takes in a `key` for random sampling, and optionally `q` and `a` as conditioning variables, 
          which can be used to condition the SDE dynamics.
        - The number of discretization steps `T_sample` controls the fidelity of the sample, with higher values providing 
          more accurate results.
    """
    def _eu_sample_fn(
        key: PRNGKeyArray, 
        q: Optional[Float[Array, "..."]] = None, 
        a: Optional[Float[Array, "..."]] = None, 
    ) -> Array: 
        return single_eu_sample_fn(
            model, sde, data_shape, key, q, a, T_sample
        ) 
    return _eu_sample_fn


def get_ode_sample_fn(
    model: eqx.Module, 
    sde: SDE, 
    data_shape: Sequence[int],
    solver: Optional[dfx.AbstractSolver] = None
) -> Callable:
    """
        Returns a callable function that implements sampling from the reverse-time stochastic differential equation 
        (SDE) using the ODE solver approach.

        Parameters:
        -----------
        `model` : `eqx.Module`
            The trained score network model used for reverse-time sampling.

        `sde` : `SDE`
            The stochastic differential equation (SDE) defining both the forward diffusion dynamics.

        `data_shape` : `Sequence[int]`
            Shape of the data to be sampled, used for constructing the prior sample.

        Returns:
        --------
        `Callable`
            A function `_ode_sample_fn` which generates a data sample given a random key `key`, and optional 
            conditioning variables `q` and `a`. It has the following signature:
            
            ```python
            def _ode_sample_fn(key: PRNGKeyArray, q: Optional[Array], a: Optional[Array]) -> Array
            ```

            This function uses `single_ode_sample_fn` to generate a sample at time `t0`.

        Example:
        --------
        ```python
        ode_sampler = get_ode_sample_fn(model, sde, data_shape=(3, 32, 32))
        sampled_data = ode_sampler(jr.PRNGKey(0))
        ```

        Notes:
        ------
        - The returned function calls `single_ode_sample_fn` to simulate the reverse diffusion process starting from 
          a prior sample at time `t1`, and evolves it to `t0` using the ODE method.
        - The sample function takes in a `key` for random sampling, and optionally `q` and `a` as conditioning variables, 
          which can be used to condition the SDE dynamics.
    """
    def _ode_sample_fn(
        key: PRNGKeyArray, 
        q: Optional[Float[Array, "..."]] = None, 
        a: Optional[Float[Array, "..."]] = None, 
    ) -> Array:
        return single_ode_sample_fn(
            model, sde, data_shape, key, q, a, solver
        )
    return _ode_sample_fn


from __future__ import annotations

from typing import Optional, Optional, List, Union, Callable

import jax
import jax.numpy as jnp
import jax.random as jr 
import equinox as eqx
import equinox as eqx
from jaxtyping import PRNGKeyArray, Array, Float, Scalar, jaxtyped
from beartype import beartype as typechecker

from einops import rearrange

typecheck = jaxtyped(typechecker=typechecker)

def identity(x):
    return x


def exists(v):
    return v is not None


def default(v, d):
    return v if exists(v) else d


def precision_cast(fn: Union[Callable, eqx.Module], x: Array) -> Array:
    return fn(x.astype(jnp.float32)).astype(x.dtype)


"""
    Diffusion Transformer (DiT)
"""


class AdaLayerNorm(eqx.Module):
    norm: eqx.nn.LayerNorm
    scale_proj: eqx.nn.Linear
    shift_proj: eqx.nn.Linear

    @typecheck
    def __init__(self, embed_dim: int, *, key: PRNGKeyArray):
        keys = jr.split(key)
        self.norm = eqx.nn.LayerNorm(embed_dim)
        self.scale_proj = eqx.nn.Linear(embed_dim, embed_dim, key=keys[0])
        self.shift_proj = eqx.nn.Linear(embed_dim, embed_dim, key=keys[1])

    @typecheck
    def __call__(self, x: Float[Array, "q"], y: Float[Array, "y"]):
        gamma = self.scale_proj(y)
        beta = self.shift_proj(y)
        return self.norm(x) * (1. + gamma) + beta


class PatchEmbedding(eqx.Module):
    patch_size: int
    proj: eqx.nn.Conv2d
    cls_token: Float[Array, "1 1 e"]
    pos_embed: Float[Array, "1 s e"]

    @typecheck
    def __init__(
        self, 
        img_size: int, 
        patch_size: int, 
        in_channels: int, 
        embed_dim: int, 
        key: PRNGKeyArray
    ):
        keys = jr.split(key, 3)
        self.patch_size = patch_size
        self.proj = eqx.nn.Conv2d(in_channels, embed_dim, kernel_size=patch_size, stride=patch_size, key=keys[0])
        self.cls_token = jr.normal(keys[1], (1, embed_dim)) # Extra 1 before
        self.pos_embed = jr.normal(keys[2], (int(img_size / patch_size) ** 2 + 1, embed_dim))

    @typecheck
    def __call__(self, x: Float[Array, "_ _ _"]) -> Float[Array, "s q"]:
        x = self.proj(x)
        x = rearrange(x, "c h w -> (h w) c") 
        x = jnp.concatenate([self.cls_token, x], axis=0) 
        x = x + self.pos_embed  
        return x


class TimestepEmbedding(eqx.Module):
    embed_dim: int
    mlp: eqx.nn.Sequential

    @typecheck
    def __init__(self, embed_dim: int, *, key: PRNGKeyArray):
        self.embed_dim = embed_dim

        keys = jr.split(key)
        self.mlp = eqx.nn.Sequential(
            [
                eqx.nn.Linear(1, embed_dim, key=keys[0]),
                eqx.nn.Lambda(jax.nn.gelu),
                eqx.nn.Linear(embed_dim, embed_dim, key=keys[1]),
            ]
        )

    @typecheck
    def __call__(self, t: Scalar) -> Float[Array, "{self.embed_dim}"]:
        return self.mlp(jnp.atleast_1d(t))


class TransformerBlock(eqx.Module):
    norm1: AdaLayerNorm
    attn: eqx.nn.MultiheadAttention
    norm2: AdaLayerNorm
    mlp: eqx.nn.Sequential

    @typecheck
    def __init__(
        self, 
        embed_dim: int, 
        n_heads: int, 
        *, 
        key: PRNGKeyArray
    ):
        keys = jr.split(key, 5)
        self.norm1 = AdaLayerNorm(embed_dim, key=keys[0])
        self.attn = eqx.nn.MultiheadAttention(n_heads, embed_dim, key=keys[1]) # NOTE: Casting in here...
        self.norm2 = AdaLayerNorm(embed_dim, key=keys[2])
        self.mlp = eqx.nn.Sequential(
            [
                eqx.nn.Linear(embed_dim, embed_dim * 4, key=keys[3]),
                eqx.nn.Lambda(jax.nn.gelu),
                eqx.nn.Linear(embed_dim * 4, embed_dim, key=keys[4])
            ]
        )

    @typecheck
    def __call__(self, x: Float[Array, "s q"], y: Float[Array, "y"]) -> Float[Array, "s q"]:
        x = precision_cast(jax.vmap(lambda x: self.norm1(x, y)), x)
        x = x + self.attn(x, x, x)
        x = precision_cast(jax.vmap(lambda x: self.norm2(x, y)), x)
        x = x + jax.vmap(self.mlp)(x)
        return x


class DiT(eqx.Module):
    img_size: int
    channels: int
    q_dim: int
    a_dim: int
    patch_embed: PatchEmbedding
    time_embed: TimestepEmbedding
    a_embed: Optional[eqx.nn.Linear]
    blocks: List[TransformerBlock]
    out_conv: eqx.nn.ConvTranspose2d
    # scaler: Optional[Scaler] = None

    @typecheck
    def __init__(
        self, 
        img_size: int, 
        patch_size: int, 
        channels: int, 
        embed_dim: int, 
        depth: int, 
        n_heads: int, 
        q_dim: Optional[int] = None, 
        a_dim: Optional[int] = None, 
        # scaler: Optional[Scaler] = None,
        *, 
        key: PRNGKeyArray
    ):
        self.img_size = img_size
        self.q_dim = q_dim
        self.a_dim = a_dim
        self.channels = channels

        keys = jr.split(key, 5)
        channels = channels + q_dim if exists(q_dim) else channels

        self.patch_embed = PatchEmbedding(
            img_size, patch_size, channels, embed_dim, key=keys[0]
        )
        self.time_embed = TimestepEmbedding(embed_dim, key=keys[1])
        
        self.a_embed = eqx.nn.Linear(a_dim, embed_dim, key=keys[2]) if exists(a_dim) else None

        block_keys = jr.split(keys[3], depth)
        self.blocks = eqx.filter_vmap(
            lambda key: TransformerBlock(embed_dim, n_heads, key=key) 
        )(block_keys)

        # self.out_conv = eqx.nn.ConvTranspose2d(
        #     embed_dim, channels, kernel_size=patch_size, stride=patch_size, key=keys[4]
        # )
        self.out_conv = eqx.nn.ConvTranspose2d(
            embed_dim, self.channels, kernel_size=patch_size, stride=patch_size, key=keys[4]
        )

        # self.scaler = scaler

    @typecheck
    def __call__(
        self, 
        t: Scalar, 
        # x: XArray, 
        # q: QArray, 
        # a: AArray,
        x: Float[Array, "{self.channels} _ _"], 
        q: Optional[Float[Array, "{self.q_dim} _ _"]] = None, 
        a: Optional[Float[Array, "{self.a_dim}"]] = None, 
        key: Optional[PRNGKeyArray] = None
    ) -> Float[Array, "_ _ _"]:

        # if exists(self.scaler):
        #     x, q, a = self.scaler.forward(x, q, a)

        x = self.patch_embed(
            jnp.concatenate([x, q]) 
            if exists(q) and exists(self.q_dim)
            else x
        )

        t_embedding = self.time_embed(t)
        if (a is not None) and (self.a_dim is not None):
            a_embedding = self.a_embed(a)
            embedding = a_embedding + t_embedding
        else:
            embedding = t_embedding

        all_params, struct = eqx.partition(self.blocks, eqx.is_array)

        def block_fn(x, params):
            block = eqx.combine(params, struct)
            x = block(x, embedding)
            return x, None

        x, _ = jax.lax.scan(block_fn, x, all_params)

        x = x[1:] # No class token 
        x = rearrange(
            x, 
            "(h w) c -> c h w", 
            h=int(self.img_size / self.patch_embed.patch_size)
        )  
        # x = self.out_conv(
        #     jnp.concatenate([x, q]) 
        #     if exists(q) and exists(self.q_dim)
        #     else x
        # )
        x = self.out_conv(x)

        return x


from typing import Optional, Sequence, Tuple
import jax
import jax.numpy as jnp
import jax.random as jr
import equinox as eqx
from jaxtyping import PRNGKeyArray, Array, Float, Scalar, jaxtyped
from beartype import beartype as typechecker

from ._sde import SDE, _get_log_prob_fn, Time, TimeFn


def get_beta_fn(beta_integral_fn: TimeFn | eqx.Module) -> TimeFn:
    """ Obtain beta function from a beta integral. """
    def _beta_fn(t: Time) -> Scalar:
        _, beta = jax.jvp(
            beta_integral_fn, 
            primals=(t,), 
            tangents=(jnp.ones_like(t),),
            has_aux=False
        )
        return beta
    return _beta_fn


class VPSDE(SDE):
    beta_integral_fn: TimeFn | eqx.Module 
    beta_fn: TimeFn 
    weight_fn: TimeFn

    @jaxtyped(typechecker=typechecker)
    def __init__(
        self, 
        beta_integral_fn: TimeFn, 
        weight_fn: Optional[TimeFn] = None, 
        dt: float = 0.1, 
        t0: float = 0., 
        t1: float = 1.
    ):
        """
            Construct a Variance Preserving SDE.
        """
        super().__init__(dt=dt, t0=t0, t1=t1)
        self.beta_integral_fn = beta_integral_fn 
        self.beta_fn = get_beta_fn(beta_integral_fn)
        self.weight_fn = weight_fn

    @jaxtyped(typechecker=typechecker)
    def sde(self, x: Float[Array, "..."], t: Time) -> Tuple[Float[Array, "..."], Scalar]:
        """ 
            dx = f(x, t) * dt + g(t) * dw 
            dx = -0.5 * beta(t) * x * dt + sqrt(beta(t)) * dw
        """
        beta_t = self.beta_fn(t)
        drift = -0.5 * beta_t * x 
        diffusion = jnp.sqrt(beta_t)
        return drift, diffusion

    @jaxtyped(typechecker=typechecker)
    def marginal_prob(self, x: Float[Array, "..."], t: Time) -> Tuple[Float[Array, "..."], Scalar]:
        """ 
            VP SDE p_t(x(t)|x(0)) is
                x(t) ~ G[x(t)|mu(x(0), t), sigma^2(t)] 
            where
                mu(x(0), t) = x(0) * exp(-0.5 * int[beta(s)])
                sigma^2(t) = I * (1 - exp(-int[beta(s)]))
        """
        beta_integral = self.beta_integral_fn(t)
        mean = jnp.exp(-0.5 * beta_integral) * x 
        std = jnp.sqrt(-jnp.expm1(-beta_integral)) 
        return mean, std

    @jaxtyped(typechecker=typechecker)
    def weight(self, t: Time, likelihood_weight: bool = False) -> Scalar:
        # likelihood weighting: above Eq 8 https://arxiv.org/pdf/2101.09258.pdf
        if self.weight_fn is not None and not likelihood_weight:
            weight = self.weight_fn(t)
        else:
            if likelihood_weight:
                weight = self.beta_fn(t) # beta(t)
            else:
                weight = -jnp.expm1(-self.beta_integral_fn(t))
        return weight

    def prior_sample(self, key: PRNGKeyArray, shape: Sequence[int]) -> Float[Array, "..."]:
        return jr.normal(key, shape)

    def prior_log_prob(self, z: Float[Array, "..."]) -> Scalar:
        return _get_log_prob_fn(scale=1.)(z)